"""
The following functions provide a convenient method
to print and access vectors by state or actuator name.
"""

from pydrake.common.containers import namedview
from pydrake.multibody.tree import (
    JointIndex,
)


def MakeNamedViewActuation(plant, view_name):
    names = [None] * plant.get_actuation_input_port().size()
    for ind in plant.GetJointActuatorIndices():
        actuator = plant.get_joint_actuator(ind)
        assert actuator.num_inputs() == 1
        names[actuator.input_start()] = actuator.name()
    return namedview(view_name, names)


# Adapted from https://github.com/RussTedrake/manipulation/blob/master/manipulation/utils.py  # noqa
# TODO(russt): consider making a version with model_instance
def MakeNamedViewPositions(
    plant, view_name, add_suffix_if_single_position=False
):
    names = [None] * plant.num_positions()
    for ind in plant.GetJointIndices():
        joint = plant.get_joint(JointIndex(ind))
        if joint.num_positions() == 1 and not add_suffix_if_single_position:
            names[joint.position_start()] = joint.name()
        else:
            for i in range(joint.num_positions()):
                names[joint.position_start() + i] = (
                    f"{joint.name()}_{joint.position_suffix(i)}"
                )
    for ind in plant.GetFloatingBaseBodies():
        body = plant.get_body(ind)
        start = body.floating_positions_start()
        for i in range(7):
            names[start + i] = (
                f"{body.name()}_{body.floating_position_suffix(i)}"
            )
    return namedview(view_name, names)


def MakeNamedViewVelocities(
    plant, view_name, add_suffix_if_single_velocity=False
):
    names = [None] * plant.num_velocities()
    for ind in plant.GetJointIndices():
        joint = plant.get_joint(JointIndex(ind))
        if joint.num_velocities() == 1 and not add_suffix_if_single_velocity:
            names[joint.velocity_start()] = joint.name()
        else:
            for i in range(joint.num_velocities()):
                names[joint.velocity_start() + i] = (
                    f"{joint.name()}_{joint.velocity_suffix(i)}"
                )
    for ind in plant.GetFloatingBaseBodies():
        body = plant.get_body(ind)
        start = body.floating_velocities_start_in_v()
        for i in range(6):
            names[start + i] = (
                f"{body.name()}_{body.floating_velocity_suffix(i)}"
            )
    return namedview(view_name, names)


def MakeNamedViewState(plant, view_name):
    # TODO(russt): this could become a nested named view, pending
    # https://github.com/RobotLocomotion/drake/pull/14973
    pview = MakeNamedViewPositions(plant, f"{view_name}_pos", True)
    vview = MakeNamedViewVelocities(plant, f"{view_name}_vel", True)
    return namedview(view_name, pview.get_fields() + vview.get_fields())



import numpy as np

from .autodiffutils import AutoDiffXd


def derivative(function, x):
    """Compute the derivative of the function evaluated at the scalar input x
    using Eigen's automatic differentiation.

    The function should be scalar-input and scalar-output.
    """
    x_ad = AutoDiffXd(value=x, size=1, offset=0)
    y_ad = function(x_ad)
    return y_ad.derivatives()[0]


def gradient(function, x):
    """Compute the gradient of the function evaluated at the vector input x
    using Eigen's automatic differentiation.

    ``function`` should be vector-input and be either a scalar output or a
    vector of size 1, where the element must be of type AutoDiffXd.
    """
    x = np.asarray(x)
    assert x.ndim == 1, "x must be a vector"
    x_ad = np.empty(x.shape, dtype=AutoDiffXd)
    for i in range(x.size):
        x_ad.flat[i] = AutoDiffXd(value=x.flat[i], size=x.size, offset=i)
    y_ad = np.asarray(function(x_ad))
    # TODO(eric.cousineau): Consider restricting this in the future to only be
    # a scalar.
    assert y_ad.size == 1 and y_ad.ndim <= 1, (
        "The output of `function` must be of a scalar or a vector of size 1"
    )
    y_ad = y_ad.reshape(())  # To scalar.
    return y_ad.item().derivatives()


def jacobian(function, x):
    """Compute the jacobian of the function evaluated at the vector input x
    using Eigen's automatic differentiation. The dimension of the jacobian will
    be one more than the output of ``function``.

    ``function`` should be vector-input, and can be any dimension output, and
    must return an array with AutoDiffXd elements.
    """
    x = np.asarray(x)
    assert x.ndim == 1, "x must be a vector"
    x_ad = np.empty(x.shape, dtype=object)
    for i in range(x.size):
        x_ad.flat[i] = AutoDiffXd(value=x.flat[i], size=x.size, offset=i)
    y_ad = np.asarray(function(x_ad))
    # An AutoDiffXd variable with empty derivatives should be treated as
    # having all zero derivatives. Checking the length of the derivatives
    # vector, and replacing it with all zeros ensures np.vstack doesn't
    # throw an error when the shapes don't line up.
    return np.vstack(
        [
            y.derivatives() if len(y.derivatives()) > 0 else np.zeros(x.size)
            for y in y_ad.flat
        ]
    ).reshape(y_ad.shape + (-1,))


# Method overloads:
# These are obviously not a complete set of mathematical functions for
# autodiff numbers. Rather, they exist just as a demonstration of *how*
# to overload individual mathematical functions to work with both
# AutoDiff and numeric inputs.


def sin(x):
    if isinstance(x, AutoDiffXd):
        return x.sin()
    else:
        return np.sin(x)


def cos(x):
    if isinstance(x, AutoDiffXd):
        return x.cos()
    else:
        return np.cos(x)



import numpy as np

from pydrake.common.value import Value
from pydrake.math import RigidTransform
from pydrake.perception import BaseField, Fields, PointCloud
from pydrake.systems.framework import LeafSystem


def _TransformPoints(points_Ci, X_CiSi):
    # Make homogeneous copy of points.
    points_h_Ci = np.vstack((points_Ci, np.ones((1, points_Ci.shape[1]))))

    return X_CiSi.dot(points_h_Ci)[:3, :]


def _TileColors(color, dim):
    # Need manual broadcasting.
    return np.tile(np.array([color]).T, (1, dim))


def _ConcatenatePointClouds(points_dict, colors_dict):
    scene_points = None
    scene_colors = None

    for id in points_dict:
        if scene_points is None:
            scene_points = points_dict[id]
        else:
            scene_points = np.hstack((points_dict[id], scene_points))

        if scene_colors is None:
            scene_colors = colors_dict[id]
        else:
            scene_colors = np.hstack((colors_dict[id], scene_colors))

    valid_indices = np.logical_not(np.isnan(scene_points))

    scene_points = scene_points[:, valid_indices[0, :]]
    scene_colors = scene_colors[:, valid_indices[0, :]]

    return scene_points, scene_colors


class PointCloudConcatenation(LeafSystem):
    """
    .. pydrake_system::

        name: PointCloudConcatenation
        input_ports:
        - point_cloud_CiSi_id0
        - X_FCi_id0
        - ...
        - point_cloud_CiSi_idN
        - X_FCi_idN
        output_ports:
        - point_cloud_FS
    """

    def __init__(self, id_list, default_rgb=[255.0, 255.0, 255.0]):
        """
        A system that takes in N point clouds of points Si in frame Ci, and N
        RigidTransforms from frame Ci to F, to put each point cloud in a common
        frame F. The system returns one point cloud combining all of the
        transformed point clouds. Each point cloud must have XYZs. RGBs are
        optional. If absent, those points will be the provided default color.

        @param id_list A list containing the string IDs of all of the point
            clouds. This is often the serial number of the camera they came
            from, such as "1" for a simulated camera or "805212060373" for a
            real camera.
        @param default_rgb A list of length 3 containing the RGB values to use
            in the absence of PointCloud.rgbs. Values should be between 0 and
            255. The default is white.
        """
        LeafSystem.__init__(self)

        self._point_cloud_ports = {}
        self._transform_ports = {}

        self._id_list = id_list

        self._default_rgb = np.array(default_rgb)

        output_fields = Fields(BaseField.kXYZs | BaseField.kRGBs)

        for id in self._id_list:
            self._point_cloud_ports[id] = self.DeclareAbstractInputPort(
                f"point_cloud_CiSi_{id}",
                Value(PointCloud(fields=output_fields)),
            )

            self._transform_ports[id] = self.DeclareAbstractInputPort(
                f"X_FCi_{id}", Value(RigidTransform.Identity())
            )

        self.DeclareAbstractOutputPort(
            "point_cloud_FS",
            lambda: Value(PointCloud(fields=output_fields)),
            self.DoCalcOutput,
        )

    def _AlignPointClouds(self, context):
        points = {}
        colors = {}

        for id in self._id_list:
            point_cloud = self.EvalAbstractInput(
                context, self._point_cloud_ports[id].get_index()
            ).get_value()
            X_CiSi = self.EvalAbstractInput(
                context, self._transform_ports[id].get_index()
            ).get_value()

            points[id] = _TransformPoints(
                point_cloud.xyzs(), X_CiSi.GetAsMatrix4()
            )

            if point_cloud.has_rgbs():
                colors[id] = point_cloud.rgbs()
            else:
                colors[id] = _TileColors(
                    self._default_rgb, point_cloud.xyzs().shape[1]
                )

        return _ConcatenatePointClouds(points, colors)

    def DoCalcOutput(self, context, output):
        scene_points, scene_colors = self._AlignPointClouds(context)

        output.get_mutable_value().resize(scene_points.shape[1])
        output.get_mutable_value().mutable_xyzs()[:] = scene_points
        output.get_mutable_value().mutable_rgbs()[:] = scene_colors



# See `ExecuteExtraPythonCode` in `pydrake_pybind.h` for usage details and
# rationale.

# ruff: noqa: F821 (undefined-name). This file is only a fragment.

from pydrake.common.value import AbstractValue as _AbstractValue


class PySerializer(SerializerInterface):
    """Provides a Python implementation of `SerializerInterface` for use
    with `LcmPublisherSystem` and `LcmSubscriberSystem` when the given
    `lcm_type` is a Python object (not a C++ object).
    """

    def __init__(self, lcm_type):
        SerializerInterface.__init__(self)
        self._lcm_type = lcm_type

    def __repr__(self):
        return f"PySerializer({self._lcm_type.__name__})"

    def CreateDefaultValue(self):
        return _AbstractValue.Make(self._lcm_type())

    def Deserialize(self, buffer, abstract_value):
        message = self._lcm_type.decode(buffer)
        abstract_value.set_value(message)

    def Serialize(self, abstract_value):
        assert isinstance(abstract_value, _AbstractValue)
        message = abstract_value.get_value()
        assert isinstance(message, self._lcm_type)
        return message.encode()


@staticmethod
def _make_lcm_subscriber(
    channel,
    lcm_type,
    lcm,
    use_cpp_serializer=False,
    *,
    wait_for_message_on_initialization_timeout=0.0,
):
    """Convenience to create an LCM subscriber system with a concrete type.

    Args:
        channel: LCM channel name.
        lcm_type: Python class generated by lcmgen.
        lcm: LCM service instance.
        use_cpp_serializer: Use C++ serializer to interface with LCM converter
            systems that are implemented in C++. LCM types must be registered
            in C++ via ``BindCppSerializer``.
        wait_for_message_on_initialization_timeout: Configures the behavior of
            initialization events (see ``System.ExecuteInitializationEvents``
            and ``Simulator.Initialize``) by specifying the number of seconds
            (wall-clock elapsed time) to wait for a new message. If this
            timeout is <= 0, initialization will copy any already-received
            messages into the Context but will not process any new messages.
            If this timeout is > 0, initialization will call
            ``lcm.HandleSubscriptions()`` until at least one message is
            received or until the timeout. Pass ∞ to wait indefinitely.
    """
    # TODO(eric.cousineau): Make `use_cpp_serializer` be kwarg-only.
    # N.B. This documentation is actually public, as it is assigned to classes
    # below as a static class method.
    if not use_cpp_serializer:
        serializer = PySerializer(lcm_type)
    else:
        serializer = _Serializer_[lcm_type]()
    return LcmSubscriberSystem(
        channel, serializer, lcm, wait_for_message_on_initialization_timeout
    )


@staticmethod
def _make_lcm_publisher(
    channel,
    lcm_type,
    lcm,
    publish_period=0.0,
    use_cpp_serializer=False,
    *,
    publish_triggers=None,
    publish_offset=0.0,
):
    """Convenience to create an LCM publisher system with a concrete type.

    Args:
        channel: LCM channel name.
        lcm_type: Python class generated by lcmgen.
        lcm: LCM service instance.
        publish_period: System's publish period (in seconds). Default is 0.
        publish_offset: System's publish offset (in seconds). Default is 0.
        use_cpp_serializer: Use C++ serializer to interface with LCM converter
            systems that are implemented in C++. LCM types must be registered
            in C++ via `BindCppSerializer`.
    """
    # TODO(eric.cousineau): Make `use_cpp_serializer` be kwarg-only.
    if not use_cpp_serializer:
        serializer = PySerializer(lcm_type)
    else:
        serializer = _Serializer_[lcm_type]()
    if publish_triggers is not None:
        return LcmPublisherSystem(
            channel=channel,
            serializer=serializer,
            lcm=lcm,
            publish_triggers=publish_triggers,
            publish_period=publish_period,
            publish_offset=publish_offset,
        )
    else:
        return LcmPublisherSystem(
            channel=channel,
            serializer=serializer,
            lcm=lcm,
            publish_period=publish_period,
            publish_offset=publish_offset,
        )


LcmSubscriberSystem.Make = _make_lcm_subscriber
LcmPublisherSystem.Make = _make_lcm_publisher



"""LCM type definitions
This file automatically generated by lcm.
DO NOT MODIFY BY HAND!!!!
"""


from io import BytesIO
import struct

import drake

class lcmt_contact_results_for_viz(object):

    __slots__ = ["timestamp", "num_point_pair_contacts", "point_pair_contact_info", "num_hydroelastic_contacts", "hydroelastic_contacts"]

    __typenames__ = ["int64_t", "int32_t", "drake.lcmt_point_pair_contact_info_for_viz", "int32_t", "drake.lcmt_hydroelastic_contact_surface_for_viz"]

    __dimensions__ = [None, None, ["num_point_pair_contacts"], None, ["num_hydroelastic_contacts"]]

    def __init__(self):
        self.timestamp = 0
        """
        in microseconds
        LCM Type: int64_t
        """

        self.num_point_pair_contacts = 0
        """ LCM Type: int32_t """
        self.point_pair_contact_info = []
        """ LCM Type: drake.lcmt_point_pair_contact_info_for_viz[num_point_pair_contacts] """
        self.num_hydroelastic_contacts = 0
        """ LCM Type: int32_t """
        self.hydroelastic_contacts = []
        """ LCM Type: drake.lcmt_hydroelastic_contact_surface_for_viz[num_hydroelastic_contacts] """

    def encode(self):
        buf = BytesIO()
        buf.write(lcmt_contact_results_for_viz._get_packed_fingerprint())
        self._encode_one(buf)
        return buf.getvalue()

    def _encode_one(self, buf):
        buf.write(struct.pack(">qi", self.timestamp, self.num_point_pair_contacts))
        for i0 in range(self.num_point_pair_contacts):
            assert self.point_pair_contact_info[i0]._get_packed_fingerprint() == drake.lcmt_point_pair_contact_info_for_viz._get_packed_fingerprint()
            self.point_pair_contact_info[i0]._encode_one(buf)
        buf.write(struct.pack(">i", self.num_hydroelastic_contacts))
        for i0 in range(self.num_hydroelastic_contacts):
            assert self.hydroelastic_contacts[i0]._get_packed_fingerprint() == drake.lcmt_hydroelastic_contact_surface_for_viz._get_packed_fingerprint()
            self.hydroelastic_contacts[i0]._encode_one(buf)

    @staticmethod
    def decode(data: bytes):
        if hasattr(data, 'read'):
            buf = data
        else:
            buf = BytesIO(data)
        if buf.read(8) != lcmt_contact_results_for_viz._get_packed_fingerprint():
            raise ValueError("Decode error")
        return lcmt_contact_results_for_viz._decode_one(buf)

    @staticmethod
    def _decode_one(buf):
        self = lcmt_contact_results_for_viz()
        self.timestamp, self.num_point_pair_contacts = struct.unpack(">qi", buf.read(12))
        self.point_pair_contact_info = []
        for i0 in range(self.num_point_pair_contacts):
            self.point_pair_contact_info.append(drake.lcmt_point_pair_contact_info_for_viz._decode_one(buf))
        self.num_hydroelastic_contacts = struct.unpack(">i", buf.read(4))[0]
        self.hydroelastic_contacts = []
        for i0 in range(self.num_hydroelastic_contacts):
            self.hydroelastic_contacts.append(drake.lcmt_hydroelastic_contact_surface_for_viz._decode_one(buf))
        return self

    @staticmethod
    def _get_hash_recursive(parents):
        if lcmt_contact_results_for_viz in parents: return 0
        newparents = parents + [lcmt_contact_results_for_viz]
        tmphash = (0x7910f7b8bb1cdb4b+ drake.lcmt_point_pair_contact_info_for_viz._get_hash_recursive(newparents)+ drake.lcmt_hydroelastic_contact_surface_for_viz._get_hash_recursive(newparents)) & 0xffffffffffffffff
        tmphash  = (((tmphash<<1)&0xffffffffffffffff) + (tmphash>>63)) & 0xffffffffffffffff
        return tmphash
    _packed_fingerprint = None

    @staticmethod
    def _get_packed_fingerprint():
        if lcmt_contact_results_for_viz._packed_fingerprint is None:
            lcmt_contact_results_for_viz._packed_fingerprint = struct.pack(">Q", lcmt_contact_results_for_viz._get_hash_recursive([]))
        return lcmt_contact_results_for_viz._packed_fingerprint

    def get_hash(self):
        """Get the LCM hash of the struct"""
        return struct.unpack(">Q", lcmt_contact_results_for_viz._get_packed_fingerprint())[0]




"""LCM type definitions
This file automatically generated by lcm.
DO NOT MODIFY BY HAND!!!!
"""


from io import BytesIO
import struct

import drake

class lcmt_point_cloud(object):
    """
    Generic point cloud with runtime typing.
    
    Modeled after PCL and ROS conventions:
    https://pointclouds.org/documentation/structpcl_1_1_p_c_l_point_cloud2.html
    https://docs.ros.org/en/api/sensor_msgs/html/msg/PointCloud2.html
    
    """

    __slots__ = ["utime", "frame_name", "width", "height", "num_fields", "fields", "flags", "point_step", "row_step", "filler_size", "filler", "data_size", "data"]

    __typenames__ = ["int64_t", "string", "int64_t", "int64_t", "int32_t", "drake.lcmt_point_cloud_field", "int64_t", "int32_t", "int64_t", "int16_t", "byte", "int64_t", "byte"]

    __dimensions__ = [None, None, None, None, None, ["num_fields"], None, None, None, None, ["filler_size"], None, ["data_size"]]

    IS_BIGENDIAN = 1
    """
    === Constants for the flags bits ===
    Set iff data is big-endian.
    """
    IS_STRICTLY_FINITE = 2
    """
    Set iff data contains only finite values.  (ROS sometimes uses "is_dense"
    to refer to this concept, but not all implementations agree on whether
    "dense" means "strictly finite" or "provides full width x height matrix
    with possibly infinite values", so we use a less ambiguous term here.)
    """

    def __init__(self):
        self.utime = 0
        """
        Timestamp in microseconds.
        LCM Type: int64_t
        """

        self.frame_name = ""
        """
        The name of a frame this data is associated with.
        LCM Type: string
        """

        self.width = 0
        """
        If the cloud is structured, this is the 2D width.
        If the cloud is unstructured, this is the number of points in the cloud.
        LCM Type: int64_t
        """

        self.height = 0
        """
        If the cloud is structured, this is the 2D height.
        If the cloud is unstructured, this is set to 1.
        LCM Type: int64_t
        """

        self.num_fields = 0
        """
        Descriptions of the fields.
        LCM Type: int32_t
        """

        self.fields = []
        """ LCM Type: drake.lcmt_point_cloud_field[num_fields] """
        self.flags = 0
        """
        Bitfield of flags describing this.  See the "IS_..." constants below.
        LCM Type: int64_t
        """

        self.point_step = 0
        """
        Number of bytes between each successive point.  This is always at least the
        sum of all field sizes, but may be larger if the data has internal padding.
        LCM Type: int32_t
        """

        self.row_step = 0
        """
        Number of bytes between each successive row.
        LCM Type: int64_t
        """

        self.filler_size = 0
        """
        Optional filler bytes to allow the raw point data to be aligned in memory.
        LCM Type: int16_t
        """

        self.filler = b""
        """ LCM Type: byte[filler_size] """
        self.data_size = 0
        """
        The raw point data.
        LCM Type: int64_t
        """

        self.data = b""
        """ LCM Type: byte[data_size] """

    def encode(self):
        buf = BytesIO()
        buf.write(lcmt_point_cloud._get_packed_fingerprint())
        self._encode_one(buf)
        return buf.getvalue()

    def _encode_one(self, buf):
        buf.write(struct.pack(">q", self.utime))
        __frame_name_encoded = self.frame_name.encode('utf-8')
        buf.write(struct.pack('>I', len(__frame_name_encoded)+1))
        buf.write(__frame_name_encoded)
        buf.write(b"\0")
        buf.write(struct.pack(">qqi", self.width, self.height, self.num_fields))
        for i0 in range(self.num_fields):
            assert self.fields[i0]._get_packed_fingerprint() == drake.lcmt_point_cloud_field._get_packed_fingerprint()
            self.fields[i0]._encode_one(buf)
        buf.write(struct.pack(">qiqh", self.flags, self.point_step, self.row_step, self.filler_size))
        buf.write(bytearray(self.filler[:self.filler_size]))
        buf.write(struct.pack(">q", self.data_size))
        buf.write(bytearray(self.data[:self.data_size]))

    @staticmethod
    def decode(data: bytes):
        if hasattr(data, 'read'):
            buf = data
        else:
            buf = BytesIO(data)
        if buf.read(8) != lcmt_point_cloud._get_packed_fingerprint():
            raise ValueError("Decode error")
        return lcmt_point_cloud._decode_one(buf)

    @staticmethod
    def _decode_one(buf):
        self = lcmt_point_cloud()
        self.utime = struct.unpack(">q", buf.read(8))[0]
        __frame_name_len = struct.unpack('>I', buf.read(4))[0]
        self.frame_name = buf.read(__frame_name_len)[:-1].decode('utf-8', 'replace')
        self.width, self.height, self.num_fields = struct.unpack(">qqi", buf.read(20))
        self.fields = []
        for i0 in range(self.num_fields):
            self.fields.append(drake.lcmt_point_cloud_field._decode_one(buf))
        self.flags, self.point_step, self.row_step, self.filler_size = struct.unpack(">qiqh", buf.read(22))
        self.filler = buf.read(self.filler_size)
        self.data_size = struct.unpack(">q", buf.read(8))[0]
        self.data = buf.read(self.data_size)
        return self

    @staticmethod
    def _get_hash_recursive(parents):
        if lcmt_point_cloud in parents: return 0
        newparents = parents + [lcmt_point_cloud]
        tmphash = (0xb996d33b5fb6120b+ drake.lcmt_point_cloud_field._get_hash_recursive(newparents)) & 0xffffffffffffffff
        tmphash  = (((tmphash<<1)&0xffffffffffffffff) + (tmphash>>63)) & 0xffffffffffffffff
        return tmphash
    _packed_fingerprint = None

    @staticmethod
    def _get_packed_fingerprint():
        if lcmt_point_cloud._packed_fingerprint is None:
            lcmt_point_cloud._packed_fingerprint = struct.pack(">Q", lcmt_point_cloud._get_hash_recursive([]))
        return lcmt_point_cloud._packed_fingerprint

    def get_hash(self):
        """Get the LCM hash of the struct"""
        return struct.unpack(">Q", lcmt_point_cloud._get_packed_fingerprint())[0]




"""LCM type definitions
This file automatically generated by lcm.
DO NOT MODIFY BY HAND!!!!
"""


from io import BytesIO
import struct

class lcmt_point_pair_contact_info_for_viz(object):

    __slots__ = ["timestamp", "body1_name", "body2_name", "contact_point", "contact_force", "normal"]

    __typenames__ = ["int64_t", "string", "string", "double", "double", "double"]

    __dimensions__ = [None, None, None, [3], [3], [3]]

    def __init__(self):
        self.timestamp = 0
        """
        TODO(edrumwri) Consider removing this data which is already present in the
        structure in which this message is stored (lcmt_contact_results_for_viz).
        In microseconds
        LCM Type: int64_t
        """

        self.body1_name = ""
        """
        Names of the colliding bodies
        LCM Type: string
        """

        self.body2_name = ""
        """ LCM Type: string """
        self.contact_point = [ 0.0 for dim0 in range(3) ]
        """
        These are all expressed in the world frame.
        LCM Type: double[3]
        """

        self.contact_force = [ 0.0 for dim0 in range(3) ]
        """ LCM Type: double[3] """
        self.normal = [ 0.0 for dim0 in range(3) ]
        """ LCM Type: double[3] """

    def encode(self):
        buf = BytesIO()
        buf.write(lcmt_point_pair_contact_info_for_viz._get_packed_fingerprint())
        self._encode_one(buf)
        return buf.getvalue()

    def _encode_one(self, buf):
        buf.write(struct.pack(">q", self.timestamp))
        __body1_name_encoded = self.body1_name.encode('utf-8')
        buf.write(struct.pack('>I', len(__body1_name_encoded)+1))
        buf.write(__body1_name_encoded)
        buf.write(b"\0")
        __body2_name_encoded = self.body2_name.encode('utf-8')
        buf.write(struct.pack('>I', len(__body2_name_encoded)+1))
        buf.write(__body2_name_encoded)
        buf.write(b"\0")
        buf.write(struct.pack('>3d', *self.contact_point[:3]))
        buf.write(struct.pack('>3d', *self.contact_force[:3]))
        buf.write(struct.pack('>3d', *self.normal[:3]))

    @staticmethod
    def decode(data: bytes):
        if hasattr(data, 'read'):
            buf = data
        else:
            buf = BytesIO(data)
        if buf.read(8) != lcmt_point_pair_contact_info_for_viz._get_packed_fingerprint():
            raise ValueError("Decode error")
        return lcmt_point_pair_contact_info_for_viz._decode_one(buf)

    @staticmethod
    def _decode_one(buf):
        self = lcmt_point_pair_contact_info_for_viz()
        self.timestamp = struct.unpack(">q", buf.read(8))[0]
        __body1_name_len = struct.unpack('>I', buf.read(4))[0]
        self.body1_name = buf.read(__body1_name_len)[:-1].decode('utf-8', 'replace')
        __body2_name_len = struct.unpack('>I', buf.read(4))[0]
        self.body2_name = buf.read(__body2_name_len)[:-1].decode('utf-8', 'replace')
        self.contact_point = struct.unpack('>3d', buf.read(24))
        self.contact_force = struct.unpack('>3d', buf.read(24))
        self.normal = struct.unpack('>3d', buf.read(24))
        return self

    @staticmethod
    def _get_hash_recursive(parents):
        if lcmt_point_pair_contact_info_for_viz in parents: return 0
        tmphash = (0x8927252e061770f4) & 0xffffffffffffffff
        tmphash  = (((tmphash<<1)&0xffffffffffffffff) + (tmphash>>63)) & 0xffffffffffffffff
        return tmphash
    _packed_fingerprint = None

    @staticmethod
    def _get_packed_fingerprint():
        if lcmt_point_pair_contact_info_for_viz._packed_fingerprint is None:
            lcmt_point_pair_contact_info_for_viz._packed_fingerprint = struct.pack(">Q", lcmt_point_pair_contact_info_for_viz._get_hash_recursive([]))
        return lcmt_point_pair_contact_info_for_viz._packed_fingerprint

    def get_hash(self):
        """Get the LCM hash of the struct"""
        return struct.unpack(">Q", lcmt_point_pair_contact_info_for_viz._get_packed_fingerprint())[0]

